import {
	Type,
	plainToInstance
} from 'class-transformer';

function stringCompare(a: string, b: string) {
	return a.localeCompare(b, undefined, {
		sensitivity: 'accent'
	}) === 0;
}

export class BingoGame {
	id: string;

	name: string;

	description: string;

	generator: 'simple' | 'srl_v5' | 'srl_v8' = 'simple';

	@Type(() => BingoGoal)
		goals: BingoGoal[] = [];

	@Type(() => BingoGroup)
		groups: BingoGroup[] = [];

	constructor(id: string, name: string, description: string) {
		this.id = id;
		this.name = name;
		this.description = description;
	}

	addGoal(goal: BingoGoal): void {
		if (this.goals.some(_goal => stringCompare(_goal.name, goal.name)))
			return;

		this.goals.push(goal);
	}

	removeGoal(goal: BingoGoal): void {
		this.goals = this.goals.filter(_goal => !stringCompare(_goal.name, goal.name));
	}

	removeGoalByName(name: string): void {
		const goal = this.goals.find(goal => stringCompare(goal.name, name));

		if (goal)
			this.removeGoal(goal);
	}

	addGroup(group: BingoGroup): void {
		if (this.groups.some(_group => stringCompare(_group.name, group.name)))
			return;

		this.groups.push(group);
	}

	removeGroup(group: BingoGroup): boolean {
		// Group has parent, don't delete
		const isParent = this.groups.find(_group => _group.parent && stringCompare(_group.parent.name, group.name));
		if (isParent) {
			console.error('This group is still a parent of:', isParent.name);

			return false;
		}

		// One or more goals still use this group, don't delete
		const isGoal = this.goals.find(goal => goal.group && stringCompare(goal.group.name, group.name));
		if (isGoal) {
			console.error('This group is still used by:', isGoal.name);

			return false;
		}

		this.groups = this.groups.filter(_group => !stringCompare(_group.name, group.name));

		return true;
	}

	removeGroupByName(name: string): boolean {
		const group = this.groups.find(group => stringCompare(group.name, name));

		return group ? this.removeGroup(group) : false;
	}

	getGoalsByTags(...tags: string[]): BingoGoal[] {
		return this.goals.filter(goal => goal.tags.some(_tag => tags.some(tag => stringCompare(_tag, tag))));
	}
}

function getRandomInt(max: number): number {
	return Math.floor(Math.random() * max);
}

export class BingoGoal {
	// id: string; // uuid-by-string maybe? https://www.npmjs.com/package/uuid-by-string

	name: string;

	group?: BingoGroup;

	tags: string[] = [];

	possible_spaces: number[] = [];

	constructor(name: string, group?: BingoGroup) {
		this.name = name;
		this.group = group;

		for (let i = 0; i < 25; i++) {
			if (getRandomInt(2) == 0)
				continue;

			this.possible_spaces.push(i);
		}
	}

	addTag(tag: string): BingoGoal {
		if (this.tags.includes(tag))
			return this;

		this.tags.push(tag);

		return this;
	}

	removeTag(tag: string): BingoGoal {
		this.tags = this.tags.filter(_tag => !stringCompare(_tag, tag));

		return this;
	}

	setGroup(group: BingoGroup) {
		this.group = group;
	}

	setPossibleSpaces(possible_spaces: number[]) {
		this.possible_spaces = possible_spaces;
	}
}

export class BingoGroup {
	name: string;

	@Type(() => BingoGroup)
		parent?: BingoGroup;

	constructor(name: string, parent?: BingoGroup) {
		this.name = name;
		this.parent = parent;
	}
}

// eslint-disable-next-line @typescript-eslint/no-unused-vars
function test(): void {
	const game = new BingoGame('yakuza_0', 'Yakuza 0', 'The funny game we used for all bingos.');
	const battle_group = new BingoGroup('Battle');
	const defeat_enemies = new BingoGroup('Defeat enemies on the street', battle_group);
	game.addGroup(battle_group);
	game.addGroup(defeat_enemies);

	const easy = new BingoGoal('Defeat 50 enemies on the street', defeat_enemies).addTag('Easy');
	const normal = new BingoGoal('Defeat 150 enemies on the street', defeat_enemies).addTag('Normal');
	const hard = new BingoGoal('Defeat 300 enemies on the street', defeat_enemies).addTag('Hard');

	game.addGoal(easy);
	game.addGoal(normal);
	game.addGoal(hard);

	// const jsonifiedInstance = instanceToPlain(game);
	// console.log(JSON.stringify(jsonifiedInstance));
	// console.log(JSON.stringify(game));

	const json_string = '{"id":"yakuza_0","name":"Yakuza 0","description":"The funny game we used for all bingos.","generator":"simple","goals":[{"name":"Defeat 50 enemies on the street","tags":["Easy"],"group":{"name":"Defeat enemies on the street","parent":{"name":"Battle"}},"possible_spaces":[]},{"name":"Defeat 150 enemies on the street","tags":["Normal"],"group":{"name":"Defeat enemies on the street","parent":{"name":"Battle"}},"possible_spaces":[]},{"name":"Defeat 300 enemies on the street","tags":["Hard"],"group":{"name":"Defeat enemies on the street","parent":{"name":"Battle"}},"possible_spaces":[]}],"groups":[{"name":"Battle"},{"name":"Defeat enemies on the street","parent":{"name":"Battle"}}]}';
	const parsed: JSON = JSON.parse(json_string);
	const parsed_game = plainToInstance(BingoGame, parsed);
	console.log(parsed_game);
	// console.log(JSON.stringify(other_game) === JSON.stringify(jsonifiedInstance));

	// parsed_game.removeGroup(defeat_enemies);
	// console.log(parsed_game.goals);
}

